import re

import pandas as pd
import plotly.graph_objects as go
import plotly.io as pio
import pytest

try:
    from pydantic.v1 import Field, ValidationError
except ImportError:  # pragma: no cov
    from pydantic import Field, ValidationError

from vizro.models import VizroBaseModel
from vizro.models.types import CapturedCallable, capture


def positional_only_function(a, /):
    pass


def var_positional_function(*args):
    pass


@pytest.mark.parametrize("function", [positional_only_function, var_positional_function])
def test_invalid_parameter_kind(function):
    with pytest.raises(
        ValueError,
        match="CapturedCallable does not accept functions with positional-only or variadic positional parameters",
    ):
        CapturedCallable(function)


def positional_or_keyword_function(a, b, c):
    return a + b + c


def keyword_only_function(a, *, b, c):
    return a + b + c


def var_keyword_function(a, **kwargs):
    return a + kwargs["b"] + kwargs["c"]


@pytest.fixture
def captured_callable(request):
    return CapturedCallable(request.param, 1, b=2)


@pytest.mark.parametrize(
    "captured_callable", [positional_or_keyword_function, keyword_only_function, var_keyword_function], indirect=True
)
class TestCallKeywordArguments:
    def test_call_provide_required_argument(self, captured_callable):
        assert captured_callable(c=3) == 1 + 2 + 3

    def test_call_override_existing_arguments(self, captured_callable):
        assert captured_callable(a=5, b=2, c=6) == 5 + 2 + 6


@pytest.mark.parametrize("captured_callable", [positional_or_keyword_function], indirect=True)
def test_call_positional_arguments(captured_callable):
    assert captured_callable(3) == 1 + 2 + 3


@pytest.mark.parametrize(
    "captured_callable", [positional_or_keyword_function, keyword_only_function, var_keyword_function], indirect=True
)
class TestDunderMethods:
    def test_getitem_known_args(self, captured_callable):
        assert captured_callable["a"] == 1
        assert captured_callable["b"] == 2

    def test_getitem_unknown_args(self, captured_callable):
        with pytest.raises(KeyError):
            captured_callable["c"]

    def test_setitem(self, captured_callable):
        captured_callable["a"] = 2
        assert captured_callable["a"] == 2


@pytest.mark.parametrize(
    "captured_callable", [positional_or_keyword_function, keyword_only_function, var_keyword_function], indirect=True
)
def test_call_positional_and_keyword_supplied(captured_callable):
    with pytest.raises(ValueError, match="does not support calling with both positional and keyword arguments"):
        captured_callable(3, c=4)


@pytest.mark.parametrize(
    "captured_callable, expectation",
    [
        (positional_or_keyword_function, pytest.raises(TypeError, match="missing 1 required positional argument: 'c'")),
        (keyword_only_function, pytest.raises(TypeError, match="missing 1 required keyword-only argument: 'c'")),
        (var_keyword_function, pytest.raises(KeyError, match="'c'")),
    ],
    indirect=["captured_callable"],
)
class TestCallMissingArgument:
    def test_call_missing_argument(self, captured_callable, expectation):
        with expectation:
            captured_callable()

    def test_call_is_memoryless(self, captured_callable, expectation):
        captured_callable(c=3)

        with expectation:
            captured_callable()


@pytest.mark.parametrize(
    "captured_callable, expectation",
    [
        (
            positional_or_keyword_function,
            pytest.raises(TypeError, match="takes 1 positional arguments but 2 were given"),
        ),
        (keyword_only_function, pytest.raises(TypeError, match="takes 0 positional arguments but 2 were given")),
        (var_keyword_function, pytest.raises(TypeError, match="takes 0 positional arguments but 2 were given")),
    ],
    indirect=["captured_callable"],
)
class TestCallPositionalArguments:
    def test_call_positional_arguments_invalid(self, captured_callable, expectation):
        with expectation:
            captured_callable(3, 4)


@pytest.mark.parametrize(
    "captured_callable, expectation",
    [
        (positional_or_keyword_function, pytest.raises(TypeError, match="got an unexpected keyword argument")),
        (keyword_only_function, pytest.raises(TypeError, match="got an unexpected keyword argument")),
        (var_keyword_function, pytest.raises(KeyError, match="'c'")),
    ],
    indirect=["captured_callable"],
)
def test_call_unknown_argument(captured_callable, expectation):
    with expectation:
        captured_callable(e=1)


def undecorated_function(a, b, c, d=4):
    return go.Figure()


@capture("action")
def decorated_action_function(a, b, c, d=4):
    return a + b + c + d


@capture("graph")
def decorated_graph_function(data_frame):
    return go.Figure()


@capture("graph")
def invalid_decorated_graph_function():
    return go.Figure()


class ModelWithAction(VizroBaseModel):
    # The import_path here makes it possible to import the above function using getattr(import_path, _target_).
    function: CapturedCallable = Field(..., import_path=__name__, mode="action")


class ModelWithGraph(VizroBaseModel):
    # The import_path here makes it possible to import the above function using getattr(import_path, _target_).
    function: CapturedCallable = Field(..., import_path=__name__, mode="graph")


class TestModelFieldPython:
    def test_decorated_action_function(self):
        model = ModelWithAction(function=decorated_action_function(a=1, b=2))
        assert model.function(c=3, d=4) == 1 + 2 + 3 + 4

    def test_decorated_graph_function(self):
        model = ModelWithGraph(function=decorated_graph_function(data_frame=pd.DataFrame()))
        assert model.function() == go.Figure()

    def test_undecorated_function(self):
        with pytest.raises(
            ValidationError,
            match=re.escape(
                "Invalid CapturedCallable. Supply a function imported from tests.unit.vizro.models.test_types or "
                "defined with decorator @capture('graph')."
            ),
        ):
            ModelWithGraph(function=undecorated_function(1, 2, 3))

    def test_wrong_mode(self):
        with pytest.raises(
            ValidationError,
            match=re.escape(
                "CapturedCallable was defined with @capture('action') rather than @capture('graph') and so "
                "is not compatible with the model."
            ),
        ):
            ModelWithGraph(function=decorated_action_function(a=1, b=2))


class TestCapture:
    def test_decorated_graph_function_missing_data_frame(self):
        with pytest.raises(ValueError, match="decorated_graph_function must supply a value to data_frame argument"):
            decorated_graph_function()

    def test_invalid_decorated_graph_function(self):
        with pytest.raises(ValueError, match="invalid_decorated_graph_function must have data_frame argument"):
            invalid_decorated_graph_function()


class TestModelFieldJSONConfig:
    def test_no_arguments(self):
        config = {"_target_": "decorated_action_function"}
        model = ModelWithAction(function=config)
        assert isinstance(model.function, CapturedCallable)
        assert model.function(a=1, b=2, c=3, d=4) == 1 + 2 + 3 + 4

    def test_some_arguments(self):
        config = {"_target_": "decorated_action_function", "a": 1, "b": 2}
        model = ModelWithAction(function=config)
        assert isinstance(model.function, CapturedCallable)
        assert model.function(c=3) == 1 + 2 + 3 + 4

    def test_all_arguments(self):
        config = {"_target_": "decorated_action_function", "a": 1, "b": 2, "c": 3}
        model = ModelWithAction(function=config)
        assert isinstance(model.function, CapturedCallable)
        assert model.function() == 1 + 2 + 3 + 4

    def test_decorated_graph_function(self):
        config = {"_target_": "decorated_graph_function", "data_frame": "data_source_name"}
        model = ModelWithGraph(function=config)
        assert model.function() == go.Figure()

    def test_no_target(self):
        config = {"a": 1, "b": 2}
        with pytest.raises(ValidationError, match="must contain the key '_target_'"):
            ModelWithGraph(function=config)

    def test_invalid_import(self):
        config = {"_target_": "invalid_function"}
        with pytest.raises(ValidationError, match="_target_=invalid_function cannot be imported"):
            ModelWithGraph(function=config)

    def test_invalid_arguments(self):
        config = {"_target_": "decorated_action_function", "e": 5}
        with pytest.raises(ValidationError, match="got an unexpected keyword argument"):
            ModelWithGraph(function=config)

    def test_undecorated_function(self):
        config = {"_target_": "undecorated_function", "a": 1, "b": 2, "c": 3}
        with pytest.raises(
            ValidationError,
            match=re.escape(
                "Invalid CapturedCallable. Supply a function imported from tests.unit.vizro.models.test_types or "
                "defined with decorator @capture('graph')."
            ),
        ):
            ModelWithGraph(function=config)

    def test_wrong_mode(self):
        config = {"_target_": "decorated_action_function", "a": 1, "b": 2}
        with pytest.raises(
            ValidationError,
            match=re.escape(
                "CapturedCallable was defined with @capture('action') rather than @capture('graph') and so "
                "is not compatible with the model."
            ),
        ):
            ModelWithGraph(function=config)

    def test_invalid_import_path(self):
        class ModelWithInvalidModule(VizroBaseModel):
            # The import_path doesn't exist.
            function: CapturedCallable = Field(..., import_path="invalid.module", mode="graph")

        config = {"_target_": "decorated_graph_function", "data_frame": "data_source_name"}

        with pytest.raises(
            ValueError, match="_target_=decorated_graph_function cannot be imported from invalid.module."
        ):
            ModelWithInvalidModule(function=config)


@capture("graph")
def decorated_graph_function_with_template(data_frame, template=None):
    fig = go.Figure()
    if template is not None:
        fig.layout.template = template
    return fig


@capture("graph")
def decorated_graph_function_crash(data_frame):
    raise RuntimeError("Crash")


# The _pio_default_template context manager resets pio.templates.default on exit, but the fixture should do this also.
# This fixtures acts as if someone has set the pio.templates.default e.g. in a Jupyter notebook.
@pytest.fixture
def set_pio_default_template(request):
    old_default = pio.templates.default
    pio.templates.default = request.param
    yield request.param
    pio.templates.default = old_default


# The expected template is always the same as the template_argument for the cases of vizro_dark and vizro_light.
# For template=None and template="plotly" the expected template is the same as the globally set default template when
# it's vizro_dark/vizro_light and vizro_dark when it's set to plotly.
@pytest.mark.parametrize(
    "set_pio_default_template, template_argument, expected_template",
    [
        ("plotly", None, "vizro_dark"),
        ("plotly", "plotly", "vizro_dark"),
        ("plotly", "vizro_dark", "vizro_dark"),
        ("plotly", "vizro_light", "vizro_light"),
        ("vizro_dark", None, "vizro_dark"),
        ("vizro_dark", "plotly", "vizro_dark"),
        ("vizro_dark", "vizro_dark", "vizro_dark"),
        ("vizro_dark", "vizro_light", "vizro_light"),
        ("vizro_light", None, "vizro_light"),
        ("vizro_light", "plotly", "vizro_light"),
        ("vizro_light", "vizro_dark", "vizro_dark"),
        ("vizro_light", "vizro_light", "vizro_light"),
    ],
    indirect=["set_pio_default_template"],
)
def test_graph_templates(set_pio_default_template, template_argument, expected_template):
    graph = decorated_graph_function_with_template(pd.DataFrame(), template=template_argument)
    assert graph.layout.template == pio.templates[expected_template]
    # The default template should be unchanged after running the captured function.
    assert pio.templates.default == set_pio_default_template


@pytest.mark.parametrize("set_pio_default_template", ["plotly", "vizro_dark", "vizro_light"], indirect=True)
def test_graph_template_crash(set_pio_default_template):
    with pytest.raises(RuntimeError, match="Crash"):
        decorated_graph_function_crash(pd.DataFrame())
    # The default template should be unchanged even if the captured function crashes.
    assert pio.templates.default == set_pio_default_template



from __future__ import annotations

from typing import TYPE_CHECKING, List, Literal

from dash import html

try:
    from pydantic.v1 import validator
except ImportError:  # pragma: no cov
    from pydantic import validator

from vizro.models import VizroBaseModel
from vizro.models._components.form import Checklist, Dropdown, RadioItems, RangeSlider, Slider
from vizro.models._layout import set_layout
from vizro.models._models_utils import _log_call, check_captured_callable, validate_min_length
from vizro.models.types import _FormComponentType

if TYPE_CHECKING:
    from vizro.models import Layout


class Form(VizroBaseModel):
    """Container for all form components to be provided to [`Page`][vizro.models.Page].

    Args:
        type (Literal["form"]): Defaults to `"form"`.
        components (List[FormComponentType]): List of components used in the form.
        layout (Layout): Defaults to `None`.

    """

    type: Literal["form"] = "form"
    components: List[_FormComponentType]
    layout: Layout = None  # type: ignore[assignment]

    # Re-used validators
    _check_captured_callable = validator("components", allow_reuse=True, each_item=True, pre=True)(
        check_captured_callable
    )
    _validate_min_length = validator("components", allow_reuse=True, always=True)(validate_min_length)
    _validate_layout = validator("layout", allow_reuse=True, always=True)(set_layout)

    @_log_call
    def pre_build(self):
        for component in self.components:
            if isinstance(component, (Slider, RangeSlider)):
                if component.min is None or component.max is None:
                    raise TypeError(f"{component.type} requires the arguments 'min' and 'max' when used within Form.")

            if isinstance(component, (Checklist, Dropdown, RadioItems)) and not component.options:
                raise TypeError(f"{component.type} requires the argument 'options' when used within Form.")

    @_log_call
    def build(self):
        components_container = self.layout.build()
        for component_idx, component in enumerate(self.components):
            components_container[f"{self.layout.id}_{component_idx}"].children = component.build()
        return html.Div(id=self.id, children=components_container)



import dash_bootstrap_components as dbc
import pandas as pd
import pytest
from asserts import assert_component_equal
from dash import html
from vizro.figures import kpi_card, kpi_card_reference

df = pd.DataFrame({"Actual": [1, 2, 3], "Reference": [2, 4, 6], "Reference Zero": [0, 0, 0]})


class TestKPICard:
    def test_kpi_card_mandatory(self):
        result = kpi_card(data_frame=df, value_column="Actual")()
        assert_component_equal(
            result,
            dbc.Card(
                [dbc.CardHeader([None, html.H4("Sum Actual", className="card-kpi-title")]), dbc.CardBody("6")],
                className="card-kpi",
            ),
        )

    def test_kpi_card_mandatory_and_optional(self):
        result = kpi_card(
            data_frame=df,
            value_column="Actual",
            icon="shopping_cart",
            title="sales",
            value_format="${value:0.2f}",
            agg_func="mean",
        )()
        assert_component_equal(
            result,
            dbc.Card(
                [
                    dbc.CardHeader(
                        [
                            html.P("shopping_cart", className="material-symbols-outlined"),
                            html.H4("sales", className="card-kpi-title"),
                        ]
                    ),
                    dbc.CardBody("$2.00"),
                ],
                className="card-kpi",
            ),
        )

    def test_value_format_missing_placeholder(self):
        with pytest.raises(
            IndexError,
            match="Replacement index 0 out of range for positional args tuple",
        ):
            kpi_card(data_frame=df, value_column="Actual", value_format="{.2f}}")()

    def test_value_format_nonexisting_placeholder(self):
        with pytest.raises(KeyError, match="reference"):
            kpi_card(data_frame=df, value_column="Actual", value_format="{reference.2f}}")()


class TestKPICardReference:
    def test_kpi_card_reference_mandatory_delta_negative(self):
        result = kpi_card_reference(data_frame=df, value_column="Actual", reference_column="Reference")()
        expected = dbc.Card(
            [
                dbc.CardHeader([None, html.H4("Sum Actual", className="card-kpi-title")]),
                dbc.CardBody("6"),
                dbc.CardFooter(
                    [
                        html.Span("arrow_circle_down", className="material-symbols-outlined"),
                        html.Span("-50.0% vs. reference (12)"),
                    ],
                    className="color-neg",
                ),
            ],
            className="card-kpi",
        )
        assert_component_equal(result, expected)

    def test_kpi_card_reference_mandatory_delta_positive(self):
        result = kpi_card_reference(data_frame=df, value_column="Reference", reference_column="Actual")()
        expected = dbc.Card(
            [
                dbc.CardHeader([None, html.H4("Sum Reference", className="card-kpi-title")]),
                dbc.CardBody("12"),
                dbc.CardFooter(
                    [
                        html.Span("arrow_circle_up", className="material-symbols-outlined"),
                        html.Span("+100.0% vs. reference (6)"),
                    ],
                    className="color-pos",
                ),
            ],
            className="card-kpi",
        )
        assert_component_equal(result, expected)

    def test_kpi_card_reference_mandatory_delta_zero(self):
        result = kpi_card_reference(data_frame=df, value_column="Actual", reference_column="Actual")()
        expected = dbc.Card(
            [
                dbc.CardHeader([None, html.H4("Sum Actual", className="card-kpi-title")]),
                dbc.CardBody("6"),
                dbc.CardFooter(
                    [
                        html.Span("arrow_circle_right", className="material-symbols-outlined"),
                        html.Span("+0.0% vs. reference (6)"),
                    ],
                    className="",
                ),
            ],
            className="card-kpi",
        )
        assert_component_equal(result, expected)

    def test_kpi_card_reference_mandatory_reference_zero(self):
        result = kpi_card_reference(data_frame=df, value_column="Actual", reference_column="Reference Zero")()
        expected = dbc.Card(
            [
                dbc.CardHeader([None, html.H4("Sum Actual", className="card-kpi-title")]),
                dbc.CardBody("6"),
                dbc.CardFooter(
                    [
                        html.Span("arrow_circle_up", className="material-symbols-outlined"),
                        html.Span("+nan% vs. reference (0)"),
                    ],
                    className="color-pos",
                ),
            ],
            className="card-kpi",
        )
        assert_component_equal(result, expected)

    def test_kpi_card_reference_mandatory_and_optional(self):
        result = kpi_card_reference(
            data_frame=df,
            value_column="Actual",
            reference_column="Reference",
            icon="shopping_cart",
            title="sales",
            # A and B distinguish the two strings to make sure the kpi_card_reference function isn't mixing them up
            value_format="A {value} is +{delta:.1f} ({delta_relative}:%) vs. {reference:.1f}",
            reference_format="B {value} is +{delta:.1f} ({delta_relative}:%) vs. {reference:.1f}",
            agg_func="mean",
        )()
        expected = dbc.Card(
            [
                dbc.CardHeader(
                    [
                        html.P("shopping_cart", className="material-symbols-outlined"),
                        html.H4("sales", className="card-kpi-title"),
                    ]
                ),
                dbc.CardBody("A 2.0 is +-2.0 (-0.5:%) vs. 4.0"),
                dbc.CardFooter(
                    [
                        html.Span("arrow_circle_down", className="material-symbols-outlined"),
                        html.Span("B 2.0 is +-2.0 (-0.5:%) vs. 4.0"),
                    ],
                    className="color-neg",
                ),
            ],
            className="card-kpi",
        )

        assert_component_equal(result, expected)

    def test_reference_format_missing_placeholder(self):
        with pytest.raises(
            IndexError,
            match="Replacement index 0 out of range for positional args tuple",
        ):
            kpi_card_reference(
                data_frame=df, value_column="Actual", reference_column="Reference", reference_format="{.2f}}"
            )()

    def test_reference_format_non_existing_placeholder(self):
        with pytest.raises(KeyError, match="delta_absolute"):
            kpi_card_reference(
                data_frame=df,
                value_column="Actual",
                reference_column="Reference",
                reference_format="{delta_absolute.2f}}",
            )()



"""Pre-defined action function "_on_page_load" to be reused in `action` parameter of VizroBaseModels."""

from typing import Any, Dict, List

from dash import ctx

from vizro.actions._actions_utils import _get_modified_page_figures
from vizro.managers._model_manager import ModelID
from vizro.models.types import capture


@capture("action")
def _on_page_load(targets: List[ModelID], **inputs: Dict[str, Any]) -> Dict[str, Any]:
    """Applies controls to charts on page once the page is opened (or refreshed).

    Args:
        targets: List of target component ids to apply on page load mechanism to
        inputs: Dict mapping action function names with their inputs e.g.
            inputs = {'filters': [], 'parameters': ['gdpPercap'], 'filter_interaction': [], 'theme_selector': True}

    Returns:
        Dict mapping target chart ids to modified figures e.g. {'my_scatter': Figure({})}

    """
    return _get_modified_page_figures(
        targets=targets,
        ctds_filter=ctx.args_grouping["external"]["filters"],
        ctds_filter_interaction=ctx.args_grouping["external"]["filter_interaction"],
        ctds_parameters=ctx.args_grouping["external"]["parameters"],
    )



from vizro.models._components.button import Button
from vizro.models._components.form.checklist import Checklist
from vizro.models._components.form.date_picker import DatePicker
from vizro.models._components.form.dropdown import Dropdown
from vizro.models._components.form.radio_items import RadioItems
from vizro.models._components.form.range_slider import RangeSlider
from vizro.models._components.form.slider import Slider

# Please keep alphabetically ordered
__all__ = ["Button", "Checklist", "DatePicker", "Dropdown", "RadioItems", "RangeSlider", "Slider"]



import pandas as pd
import pytest
import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro


@pytest.fixture
def gapminder_2007(gapminder):
    return gapminder.query("year == 2007")


@pytest.fixture
def iris():
    return px.data.iris()


@pytest.fixture
def gapminder_dynamic_first_n_last_n_function(gapminder):
    return lambda first_n=None, last_n=None: (
        pd.concat([gapminder[:first_n], gapminder[-last_n:]])
        if last_n
        else gapminder[:first_n] if first_n else gapminder
    )


@pytest.fixture
def box_params():
    return {"x": "continent", "y": "lifeExp", "custom_data": ["continent"]}


@pytest.fixture
def box_chart(gapminder_2007, box_params):
    return px.box(gapminder_2007, **box_params).update_layout(margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24)


@pytest.fixture
def box_chart_dynamic_data_frame(box_params):
    return px.box("gapminder_dynamic_first_n_last_n", **box_params).update_layout(
        margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24
    )


@pytest.fixture
def scatter_params():
    return {"x": "gdpPercap", "y": "lifeExp"}


@pytest.fixture
def scatter_chart(gapminder_2007, scatter_params):
    return px.scatter(gapminder_2007, **scatter_params).update_layout(
        margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24
    )


@pytest.fixture
def scatter_matrix_params():
    return {"dimensions": ["sepal_width", "sepal_length", "petal_width", "petal_length"]}


@pytest.fixture
def scatter_matrix_chart(iris, scatter_matrix_params):
    return px.scatter_matrix(iris, **scatter_matrix_params).update_layout(
        margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24
    )


@pytest.fixture
def scatter_chart_dynamic_data_frame(scatter_params):
    return px.scatter("gapminder_dynamic_first_n_last_n", **scatter_params).update_layout(
        margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24
    )


@pytest.fixture
def target_scatter_filtered_continent(request, gapminder_2007, scatter_params):
    continent = request.param
    data = gapminder_2007[gapminder_2007["continent"].isin(continent)]
    return px.scatter(data, **scatter_params).update_layout(margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24)


@pytest.fixture
def target_box_filtered_continent(request, gapminder_2007, box_params):
    continent = request.param
    data = gapminder_2007[gapminder_2007["continent"].isin(continent)]
    return px.box(data, **box_params).update_layout(margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24)


@pytest.fixture
def managers_one_page_two_graphs_one_button(box_chart, scatter_chart):
    """Instantiates a simple model_manager and data_manager with a page, two graph models and the button component."""
    vm.Page(
        id="test_page",
        title="My first dashboard",
        components=[
            vm.Graph(id="box_chart", figure=box_chart),
            vm.Graph(id="scatter_chart", figure=scatter_chart),
            vm.Button(id="button"),
        ],
    )
    Vizro._pre_build()


@pytest.fixture
def managers_one_page_two_graphs_with_dynamic_data(box_chart_dynamic_data_frame, scatter_chart_dynamic_data_frame):
    """Instantiates a simple model_manager and data_manager with a page, two graph models and the button component."""
    vm.Page(
        id="test_page",
        title="My first dashboard",
        components=[
            vm.Graph(id="box_chart", figure=box_chart_dynamic_data_frame),
            vm.Graph(id="scatter_chart", figure=scatter_chart_dynamic_data_frame),
            vm.Button(id="button"),
        ],
    )
    Vizro._pre_build()


@pytest.fixture
def managers_one_page_two_graphs_one_table_one_aggrid_one_button(
    box_chart, scatter_chart, dash_data_table_with_id, ag_grid_with_id
):
    """Instantiates a simple model_manager and data_manager with: page, graphs, table, aggrid and button component."""
    vm.Page(
        id="test_page",
        title="My first dashboard",
        components=[
            vm.Graph(id="box_chart", figure=box_chart),
            vm.Graph(id="scatter_chart", figure=scatter_chart),
            vm.Table(id="vizro_table", figure=dash_data_table_with_id),
            vm.AgGrid(id="ag_grid", figure=ag_grid_with_id),
            vm.Button(id="button"),
        ],
    )
    Vizro._pre_build()


@pytest.fixture
def managers_one_page_one_graph_with_dict_param_input(scatter_matrix_chart):
    """Instantiates a model_manager and data_manager with a page and a graph that requires a list input."""
    vm.Page(
        id="test_page",
        title="My first dashboard",
        components=[
            vm.Graph(id="scatter_matrix_chart", figure=scatter_matrix_chart),
        ],
    )
    Vizro._pre_build()



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

gapminder = px.data.gapminder()

page = vm.Page(
    title="Treemap",
    components=[
        vm.Graph(
            figure=px.treemap(
                gapminder.query("year == 2007"),
                path=[px.Constant("world"), "continent", "country"],
                values="pop",
                color="lifeExp",
            )
        ),
    ],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

gapminder = px.data.gapminder()


page = vm.Page(
    title="Ordered column",
    components=[
        vm.Graph(
            figure=px.bar(
                gapminder.query(
                    "year == 2007 and country.isin(['United States', 'Pakistan', 'India', 'China', 'Indonesia'])"
                ).sort_values("pop"),
                y="pop",
                x="country",
            )
        )
    ],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



import logging
import uuid

from vizro.managers import data_manager

logger = logging.getLogger(__name__)


def _process_callable_data_frame(captured_callable):
    # Possibly all this validator's functionality should move into CapturedCallable (or a subclass of it) in the
    # future. This would mean that data is added to the data manager outside the context of a dashboard though,
    # which might not be desirable.
    data_frame = captured_callable["data_frame"]

    if isinstance(data_frame, str):
        # Named data source, which could be dynamic or static. This means px.scatter("iris") from the Python API and
        # specification of "data_frame": "iris" through JSON. In these cases, data already exists in the data manager.
        return captured_callable

    # Unnamed data source, which must be a pd.DataFrame and hence static data. This means px.scatter(pd.DataFrame())
    # and is only possible from the Python API. Extract dataframe from the captured function and put it into the
    # data manager.
    # Unlike with model_manager, it doesn't matter if the random seed is different across workers here. So long as
    # we always fetch static data from the data manager by going through the appropriate Figure component, the right
    # data source name will be fetched. It also doesn't matter if multiple Figures with the same underlying data
    # each have their own entry in the data manager, since the underlying pd.DataFrame will still be the same and
    # not copied into each one, so no memory is wasted.
    # Replace the "data_frame" argument in the captured callable with the data_source_name for consistency with
    # dynamic data and to save memory. This way we always access data via the same interface regardless of whether it's
    # static or dynamic.
    data_source_name = str(uuid.uuid4())
    data_manager[data_source_name] = data_frame
    captured_callable["data_frame"] = data_source_name

    return captured_callable
